#!/usr/bin/env node
import TurndownService from "turndown";
import type { ConversationTree, Post, TopicData } from "./types";
/**
 * Sanitizes HTML content using DOMPurify with allowed tags for markdown conversion
 */
export declare function sanitize(html: string): string;
/**
 * Generates a safe file path with sanitized filename and path traversal protection
 */
export declare function getSafeOutputPath(targetDir: string, topicId: string, slug: string): string;
/**
 * Checks if a URL points to a user avatar image
 */
export declare function isAvatarUrl(url: string): boolean;
/**
 * Checks if a DOM node is inside an onebox embed
 */
export declare function isInsideOnebox(node: Element): boolean;
/**
 * Formats a date as ISO date string (YYYY-MM-DD)
 */
export declare function formatDateISO(date: Date | string): string;
/**
 * Escapes quotes in strings for YAML compatibility
 */
export declare function escapeYamlString(str: string): string;
/**
 * Converts emoji images to Unicode symbols using node-emoji
 */
export declare function convertEmojiImage(alt: string, src: string): string;
/**
 * Parses a Discourse URL to extract base URL and topic ID
 */
export declare function parseDiscourseUrl(fullUrl: string): {
    baseUrl: string;
    topicId: string;
} | null;
/**
 * Creates a TurndownService configured with Discourse-specific conversion rules
 */
export declare function createTurndownService(): TurndownService;
/**
 * Post-processes markdown to group consecutive images into gallery layout
 */
export declare function processImageGalleries(markdown: string): string;
/**
 * Builds a hierarchical conversation tree from flat post data
 */
export declare function buildConversationTree(posts: Post[]): ConversationTree;
/**
 * Generates markdown for external links from post link_counts
 */
export declare function renderExternalLinks(post: Post): string;
/**
 * Fetches all posts for a topic, including paginated content
 */
export declare function fetchAllPosts(baseUrl: string, topicId: string, initialTopicData: TopicData): Promise<Post[]>;
/**
 * Generates YAML frontmatter for Obsidian compatibility
 */
export declare function generateYamlFrontmatter(topicData: TopicData, originalPost: Post, sourceUrl: string, originalPostMarkdown: string): string[];
/**
 * Renders posts and replies as nested markdown blockquotes
 */
export declare function renderPostTree(posts: Post[], childrenMap: Map<number, Post[]>, baseUrl: string, topicId: string, turndownService: TurndownService): string[];
/**
 * Saves markdown content to file with safe path handling
 */
export declare function saveMarkdownFile(content: string, targetDir: string, topicId: string, slug: string): Promise<string>;
/**
 * Orchestrates the complete markdown content generation pipeline
 */
export declare function generateMarkdownContent(topicData: TopicData, allPosts: Post[], sourceUrl: string, baseUrl: string, topicId: string): string;
/**
 * Main application entry point - orchestrates the complete scraping workflow
 */
export declare function main(sourceUrl: string, targetDir: string): Promise<void>;
